@@ -33,6 +33,7 @@
 #include "http.h"
 #include "io.h"
 #include "str.h"
+#include "mem.h"
 #include "tcp.h"
 #include "res.h"
 #include "utils.h"
@@ -44,7 +45,7 @@ int quiet = 0;
 char machine_readable = 0;
 char nagios_mode = 0;
 
-char *last_error;
+char last_error[ERROR_BUFFER_SIZE];
 
 void version(void)
 {
@@ -64,8 +65,16 @@ void usage(void)
 	fprintf(stderr, "-i interval	delay between each connect\n");
 	fprintf(stderr, "-t timeout	timeout (default: 30s)\n");
 	fprintf(stderr, "-s		show statuscodes\n");
+	fprintf(stderr, "-G		do a GET request instead of HEAD (read the\n");
+	fprintf(stderr, "		contents of the page as well)\n");
+	fprintf(stderr, "-b		show transfer speed in KB/s (use with -G)\n");
+	fprintf(stderr, "-B		like -b but use compression if available\n");
+	fprintf(stderr, "-L x		limit the amount of data transferred (for -b)\n");
+	fprintf(stderr, "		to 'x' (in bytes)\n");
+	fprintf(stderr, "-X		show the number of KB transferred (for -b)\n");
 #ifndef NO_SSL
 	fprintf(stderr, "-l		connect using SSL\n");
+	fprintf(stderr, "-z		show fingerprint (SSL)\n");
 #endif
 	fprintf(stderr, "-f		flood connect (no delays)\n");
 	fprintf(stderr, "-m		give machine parseable output (see\n");
@@ -134,109 +143,150 @@ int main(int argc, char *argv[])
 	double nagios_warn=0.0, nagios_crit=0.0;
 	int nagios_exit_code = 2;
 	double avg_httping_time = -1.0;
+	int get_instead_of_head = 0;
+	char *buffer;
+	int page_size = sysconf(_SC_PAGESIZE);
+	char show_Bps = 0, ask_compression = 0;
+	int Bps_min = 1 << 30, Bps_max = 0;
+	long long int Bps_avg = 0;
+	int Bps_limit = -1;
+	char show_bytes_xfer = 0, show_fp = 0;
+	int fd = -1;
+	SSL *ssl_h = NULL;
+	BIO *s_bio;
 
-        while((c = getopt(argc, argv, "g:h:p:c:i:x:t:o:e:flqsmV?I:R:rn:N:")) != -1)
-        {
-                switch(c)
-                {
-		case 'e':
-			err_str = optarg;
-			break;
+	if (page_size == -1)
+		page_size = 4096;
 
-		case 'o':
-			ok_str = optarg;
-			break;
+	buffer = (char *)mymalloc(page_size, "receive buffer");
 
-		case 'x':
-			proxy = optarg;
-			break;
+	while((c = getopt(argc, argv, "XL:bBg:h:p:c:i:Gx:t:o:e:flqsmV?I:R:rn:N:z")) != -1)
+	{
+		switch(c)
+		{
+			case 'z':
+				show_fp = 1;
+				break;
 
-		case 'g':
-			get = optarg;
-			break;
+			case 'X':
+				show_bytes_xfer = 1;
+				break;
 
-		case 'r':
-			resolve_once = 1;
-			break;
+			case 'L':
+				Bps_limit = atoi(optarg);
+				break;
 
-		case 'h':
-			hostname = optarg;
-			break;
+			case 'B':
+				show_Bps = 1;
+				ask_compression = 1;
+				break;
 
-		case 'p':
-			portnr = atoi(optarg);
-			break;
+			case 'b':
+				show_Bps = 1;
+				break;
 
-		case 'c':
-			count = atoi(optarg);
-			break;
+			case 'e':
+				err_str = optarg;
+				break;
 
-		case 'i':
-			wait = atoi(optarg);
-			break;
+			case 'o':
+				ok_str = optarg;
+				break;
 
-		case 't':
-			timeout = atoi(optarg);
-			break;
+			case 'x':
+				proxy = optarg;
+				break;
 
-		case 'I':
-			useragent = optarg;
-			break;
+			case 'g':
+				get = optarg;
+				break;
 
-		case 'R':
-			referer = optarg;
-			break;
+			case 'r':
+				resolve_once = 1;
+				break;
 
-		case 'f':
-			wait = 0;
-			break;
+			case 'h':
+				hostname = optarg;
+				break;
+
+			case 'p':
+				portnr = atoi(optarg);
+				break;
+
+			case 'c':
+				count = atoi(optarg);
+				break;
+
+			case 'i':
+				wait = atoi(optarg);
+				break;
+
+			case 't':
+				timeout = atoi(optarg);
+				break;
+
+			case 'I':
+				useragent = optarg;
+				break;
+
+			case 'R':
+				referer = optarg;
+				break;
+
+			case 'f':
+				wait = 0;
+				break;
+
+			case 'G':
+				get_instead_of_head = 1;
+				break;
 
 #ifndef NO_SSL
-		case 'l':
-			use_ssl = 1;
-			break;
+			case 'l':
+				use_ssl = 1;
+				break;
 #endif
 
-		case 'm':
-			machine_readable = 1;
-			break;
+			case 'm':
+				machine_readable = 1;
+				break;
 
-		case 'q':
-			quiet = 1;
-			break;
+			case 'q':
+				quiet = 1;
+				break;
 
-		case 's':
-			show_statuscodes = 1;
-			break;
+			case 's':
+				show_statuscodes = 1;
+				break;
 
-		case 'V':
-			version();
-			return 0;
+			case 'V':
+				version();
+				return 0;
 
-		case 'n':
-		{
-			char *dummy = strchr(optarg, ',');
-			if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
-			nagios_mode = 1;
-			if (!dummy)
-				error_exit("-n: missing parameter\n");
-			nagios_warn = atof(optarg);
-			nagios_crit = atof(dummy + 1);
-		} break;
-		case 'N':
-			if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
-			nagios_mode = 2;
-			nagios_exit_code = atoi(optarg);
-			break;
+			case 'n':
+				{
+					char *dummy = strchr(optarg, ',');
+					if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
+					nagios_mode = 1;
+					if (!dummy)
+						error_exit("-n: missing parameter\n");
+					nagios_warn = atof(optarg);
+					nagios_crit = atof(dummy + 1);
+				} break;
+			case 'N':
+				if (nagios_mode) error_exit("-n and -N are mutual exclusive\n");
+				nagios_mode = 2;
+				nagios_exit_code = atoi(optarg);
+				break;
 
-		case '?':
-		default:
-			usage();
-			return 1;
+			case '?':
+			default:
+				version();
+				usage();
+				return 1;
 		}
 	}
 
-	last_error = mymalloc(4096, "error buffer");
 	last_error[0] = 0x00;
 
 #ifndef NO_SSL
@@ -244,10 +294,13 @@ int main(int argc, char *argv[])
 		portnr = 443;
 #endif
 
+	if (!get_instead_of_head && show_Bps)
+		error_exit("-b/-B can only be used when also using -G\n");
+
 	if (get != NULL && hostname == NULL)
 	{
 		char *slash, *colon;
-		char *getcopy = mystrdup(get);
+		char *getcopy = mystrdup(get, "get request");
 		char *http_string = "http://";
 		int http_string_len = 7;
 
@@ -324,20 +377,23 @@ int main(int argc, char *argv[])
 		client_ctx = initialize_ctx();
 		if (!client_ctx)
 		{
-			sprintf(last_error, "problem creating SSL context\n");
+			snprintf(last_error, ERROR_BUFFER_SIZE, "problem creating SSL context\n");
 			goto error_exit;
 		}
 	}
 #endif
 
 	request = mymalloc(strlen(get) + 4096, "request");
-	sprintf(request, "HEAD %s HTTP/1.0\r\n", get);
+	sprintf(request, "%s %s HTTP/1.0\r\n", get_instead_of_head?"GET":"HEAD", get);
 	if (useragent)
 		sprintf(&request[strlen(request)], "User-Agent: %s\r\n", useragent);
 	else
 		sprintf(&request[strlen(request)], "User-Agent: HTTPing v" VERSION "\r\n");
+	sprintf(&request[strlen(request)], "Host: %s\r\n", hostname);
 	if (referer)
 		sprintf(&request[strlen(request)], "Referer: %s\r\n", referer);
+	if (ask_compression)
+		sprintf(&request[strlen(request)], "Accept-Encoding: gzip,deflate\r\n");
 	strcat(request, "\r\n");
 	req_len = strlen(request);
 
@@ -356,7 +412,7 @@ int main(int argc, char *argv[])
 	{
 		memset(&addr, 0x00, sizeof(addr));
 
-		if (resolve_host(host, &addr) == 0)
+		if (resolve_host(host, &addr) == -1)
 		{
 			err++;
 			emit_error();
@@ -368,25 +424,18 @@ int main(int argc, char *argv[])
 	{
 		double ms;
 		double dstart, dend;
-		struct timeval start, end;
-        	struct timezone tz;
 		char *reply;
-		int fd;
+		int Bps = 0;
+		char is_compressed = 0;
+		long long int bytes_transferred = 0;
 
-		if (gettimeofday(&start, &tz) == -1)
-		{
-			sprintf(last_error, "gettimeofday failed: %s\n", strerror(errno));
-			ok = 0;
-			emit_error();
-			break;
-		}
+		dstart = get_ts();
 
 		for(;;)
 		{
+			char *fp = NULL;
 			int rc;
 			char *sc = NULL;
-			SSL *ssl_h = NULL;
-			BIO *s_bio;
 
 			curncount++;
 
@@ -414,6 +463,10 @@ int main(int argc, char *argv[])
 
 			if (fd >= 0)
 			{
+				/* set socket to low latency */
+				if (set_tcp_low_latency(fd) == -1)
+					break;
+
 				/* set fd blocking */
 				if (set_fd_blocking(fd) == -1)
 					break;
@@ -436,7 +489,7 @@ int main(int argc, char *argv[])
 			if (fd < 0)
 			{
 				if (fd == -2)
-					sprintf(last_error, "timeout connecting to host\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "timeout connecting to host\n");
 
 				emit_error();
 
@@ -454,13 +507,13 @@ int main(int argc, char *argv[])
 			if (rc != req_len)
 			{
 				if (rc == -1)
-					sprintf(last_error, "error sending request to host\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "error sending request to host\n");
 				else if (rc == -2)
-					sprintf(last_error, "timeout sending to host\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "timeout sending to host\n");
 				else if (rc == -3)
 				{/* ^C */}
 				else if (rc == 0)
-					sprintf(last_error, "connection prematurely closed by peer\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "connection prematurely closed by peer\n");
 
 				emit_error();
 
@@ -472,25 +525,38 @@ int main(int argc, char *argv[])
 
 			rc = get_HTTP_headers(fd, ssl_h, &reply, timeout);
 
-			if (show_statuscodes || machine_readable)
+			if ((show_statuscodes || machine_readable) && reply != NULL)
 			{
-				if (reply)
+				/* statuscode is in first line behind
+				 * 'HTTP/1.x'
+				 */
+				char *dummy = strchr(reply, ' ');
+
+				if (dummy)
 				{
-					/* statuscode is in first line behind
-					 * 'HTTP/1.x'
+					sc = strdup(dummy + 1);
+
+					/* lines are normally terminated with a
+					 * LF
 					 */
-					char *dummy = strchr(reply, ' ');
-
+					dummy = strchr(sc, '\n');
 					if (dummy)
-					{
-						sc = strdup(dummy + 1);
+						*dummy = 0x00;
+				}
+			}
 
-						/* lines are normally terminated with a
-						 * LF
-						 */
-						dummy = strchr(sc, '\n');
-						if (dummy)
-							*dummy = 0x00;
+			if (ask_compression && reply != NULL)
+			{
+				char *encoding = strstr(reply, "\nContent-Encoding:");
+				if (encoding)
+				{
+					char *dummy = strchr(encoding + 1, '\n');
+					if (dummy) *dummy = 0x00;
+
+					if (strstr(encoding, "gzip") == 0 ||
+							strstr(encoding, "deflate") == 0)
+					{
+						is_compressed = 1;
 					}
 				}
 			}
@@ -500,9 +566,9 @@ int main(int argc, char *argv[])
 			if (rc < 0)
 			{
 				if (rc == RC_SHORTREAD)
-					sprintf(last_error, "error receiving reply from host\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "error receiving reply from host\n");
 				else if (rc == RC_TIMEOUT)
-					sprintf(last_error, "timeout receiving reply from host\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "timeout receiving reply from host\n");
 
 				emit_error();
 
@@ -514,28 +580,55 @@ int main(int argc, char *argv[])
 
 			ok++;
 
-			if (gettimeofday(&end, &tz) == -1)
+			if (get_instead_of_head)
 			{
-				sprintf(last_error, "gettimeofday failed: %s\n", strerror(errno));
-				ok = 0;
-				emit_error();
-				break;
+				double dl_start = get_ts(), dl_end;
+
+				for(;;)
+				{
+					int rc = read(fd, buffer, page_size);
+
+					if (rc == -1)
+					{
+						if (errno != EINTR && errno != EAGAIN)
+							error_exit("read failed");
+					}
+					else if (rc == 0)
+						break;
+
+					bytes_transferred += rc;
+
+					if (Bps_limit != -1 && bytes_transferred >= Bps_limit)
+						break;
+				}
+
+				dl_end = get_ts();
+
+				Bps = bytes_transferred / max(dl_end - dl_start, 0.000001);
+				Bps_min = min(Bps_min, Bps);
+				Bps_max = max(Bps_max, Bps);
+				Bps_avg += Bps;
 			}
 
+			dend = get_ts();
+
 #ifndef NO_SSL
 			if (use_ssl)
 			{
+				if (show_fp && ssl_h != NULL)
+				{
+					fp = get_fingerprint(ssl_h);
+				}
+
 				if (close_ssl_connection(ssl_h, fd) == -1)
 				{
-					sprintf(last_error, "error shutting down ssl\n");
+					snprintf(last_error, ERROR_BUFFER_SIZE, "error shutting down ssl\n");
 					emit_error();
 				}
 			}
 #endif
 			close(fd);
 
-			dstart = (((double)start.tv_sec) + ((double)start.tv_usec)/1000000.0);
-			dend = (((double)end.tv_sec) + ((double)end.tv_usec)/1000000.0);
 			ms = (dend - dstart) * 1000.0;
 			avg += ms;
 			min = min > ms ? ms : min;
@@ -566,7 +659,29 @@ int main(int argc, char *argv[])
 			}
 			else if (!quiet && !nagios_mode)
 			{
-				printf("connected to %s:%d, seq=%d time=%.2f ms %s\n", hostname, portnr, curncount-1, ms, sc?sc:"");
+				printf("connected to %s:%d, seq=%d time=%.2f ms %s", hostname, portnr, curncount-1, ms, sc?sc:"");
+
+				if (show_Bps)
+				{
+					printf(" %dKB/s", Bps / 1024);
+					if (show_bytes_xfer)
+						printf(" %dKB", (int)(bytes_transferred / 1024));
+					if (ask_compression)
+					{
+						printf(" (");
+						if (!is_compressed)
+							printf("not ");
+						printf("compressed)");
+					}
+				}
+
+				if (use_ssl && show_fp && fp != NULL)
+				{
+					printf(" %s", fp);
+					free(fp);
+				}
+
+				printf("\n");
 			}
 
 			free(sc);
@@ -589,7 +704,12 @@ int main(int argc, char *argv[])
 			printf("%d connects, %d ok, %3.2f%% failed\n", curncount, ok, (((double)err) / ((double)count)) * 100.0);
 
 		if (ok > 0)
+		{
 			printf("round-trip min/avg/max = %.1f/%.1f/%.1f ms\n", min, avg_httping_time, max);
+
+			if (show_Bps)
+				printf("Transfer speed: min/avg/max = %d/%d/%d KB\n", Bps_min / 1024, (int)(Bps_avg / ok) / 1024, Bps_max / 1024);
+		}
 	}
 
 error_exit:
