@@ -19,6 +19,9 @@
 #include <sys/socket.h>
 #include <openssl/ssl.h>
 #include <openssl/bio.h>
+#include <openssl/evp.h>
+#include <openssl/x509.h>
+#include <openssl/md5.h>
 
 #include "mssl.h"
 
@@ -166,3 +169,39 @@ SSL_CTX * initialize_ctx(void)
 
 	return SSL_CTX_new(meth);
 }
+
+char * get_fingerprint(SSL *ssl_h)
+{
+	char *string = NULL;
+
+	unsigned char fp_digest[EVP_MAX_MD_SIZE];
+	X509 *x509_data = SSL_get_peer_certificate(ssl_h);
+
+	if (x509_data)
+	{
+		unsigned int fp_digest_size = sizeof(fp_digest);
+
+		memset(fp_digest, 0x00, sizeof(fp_digest));
+
+		if (X509_digest(x509_data, EVP_md5(), fp_digest, &fp_digest_size))
+		{
+			string = (char *)malloc(MD5_DIGEST_LENGTH * 3 + 1);
+			if (string)
+			{
+				int loop, pos =0;
+
+				for(loop=0; loop<MD5_DIGEST_LENGTH; loop++)
+				{
+					if (loop)
+						pos += sprintf(&string[pos], ":%02x", fp_digest[loop]);
+					else
+						pos = sprintf(&string[pos], "%02x", fp_digest[loop]);
+				}
+			}
+		}
+
+		X509_free(x509_data);
+	}
+
+	return string;
+}
